@@ -1,7 +1,6 @@
 =head1 NAME
 
-Tk::SMListbox - Sortable Multicolumn Listbox with arrows in headers 
-    indicating sort order.
+Tk::SMListbox - Sortable Multicolumn Listbox with arrows in headers indicating sort order.
 
 =head1 AUTHOR
 
@@ -23,6 +22,10 @@ Tk::SMListbox is a derivitive of Tk::MListbox that adds optional tiny up and
 down arrow graphics to columns in the SMListbox that the data is 
 currently sorted on indicating ascending and/or descending order.
 
+It also now supports "-state => 'disabled'", but note that the state must not 
+be disabled when inserting rows (inserts silently fail)!  L<Tk::HMListbox> is 
+not subject to this restriction.
+
 Sorting is done by clicking on one of the column headings in the 
 widget. The first click will sort the data with the selected column as 
 key, a new click will reverse the sort order.
@@ -73,10 +76,11 @@ my $table = $w->Scrolled('SMListbox'
 
 =head1 STANDARD OPTIONS
 
-B<-background> B<-foreground> B<-relief> B<-takefocus> B<-borderwidth>   B<-heigh>
-B<-selectbackground> B<-cursor> B<-highlightbackground> B<-selectborderwidth>
-B<-xscrollcommand> B<-exportselection> B<-highlightcolor> B<-selectforeground>
-B<-yscrollcommand> B<-font> B<-highlightthickness> B<-setgrid>
+B<-background> B<-borderwidth> B<-cursor> B<-disabledforeground> 
+B<-exportselection> B<-font> B<-foreground> B<-height> B<-highlightbackground> 
+B<-highlightcolor> B<-highlightthickness> B<-relief> B<-selectbackground> 
+B<-selectborderwidth> B<-selectforeground> B<-setgrid> B<-state> B<-takefocus> 
+B<-width> B<-xscrollcommand> B<-yscrollcommand>
 
 =head1 WIDGET SPECIFIC OPTIONS
 
@@ -173,6 +177,15 @@ Default:  B<black>.
 Note that you can also specify -separatorcolor on a column
 by column basis. See the B<COLUMNS> section below.
 
+=item B<-scrollbarfocus> => I<value>
+
+Specifies which scrollbars (if any) are allowed to take keyboard 
+focus via the tab-circulate order:  Valid values are:  
+B<0> (no scrollbars take focus), B<1> or B<"xy"> (all scrollbars take focus, 
+B<"x"> (only horizontal scrollbar takes focus), or B<"y"> (only vertical 
+scrollbar takes focus).  Default is B<""> (Allow all to take focus based 
+on Tk's default focusing model).
+
 =item B<-separatorwidth> => I<integer>
 
 Specifies the width in pixels of the separator lines 
@@ -188,9 +201,30 @@ A value of B<1> indicates that it is okay for the user to sort
 the data by clicking column headings. B<0> disables this function.
 Default:  B<1>.
 
-Note that you can also specify -sortable on a column
+Note that you can also specify B<-sortable> on a column
 by column basis. See I<COLUMNS> below.
 
+=item B<-state>
+
+Specifies one of two states for the widget: B<"normal"> or B<"disabled">.
+If the widget is disabled then user can interact with the widget in a very 
+limited (read-only) way.  Items may not be selected, but currently, the 
+list can be resorted or the columns potentially expanded for easier viewing.  
+Items are drawn in the B<-disabledforeground> color, and selection
+cannot be modified and is not shown (though selection information is
+retained).
+
+=item B<-takefocus>
+
+There are actually three different focusing options:  Specify B<1> to both 
+allow the widget to take keyboard focus itself and to enable grabbing the 
+keyboard focus when a user clicks on a row in the listbox.  Specify B<''> 
+to allow the widget to take focus (via the <TAB> circulate order) but do 
+not grab the focus when a user clicks on (selects) a row.  This is the 
+default focusing model.  Specify B<0> to not allow the widget to receive 
+the keyboard focus.
+Default:  B<''>.
+
 =back
 
 =head1 WIDGET SPECIFIC, COLUMN-SPECIFIC OPTIONS
@@ -241,6 +275,9 @@ Each column has its own set of options which might be passed to
 SMListbox::configure(-columns), SMListbox::insert(),
 SMListbox::columnConfigure() or SMLColumn::configure().
 
+NOTE:  Unlike HListbox widgets, Listbox widgets will IGNORE data-insertions 
+if the current state is B<'disabled'>!
+
 The following code snippets are all equal:
 
 1.  $ml=$mw->SMListbox(-columns=>[[-text=>'Heading1',
@@ -734,8 +771,6 @@ L<Tk::MListbox> L<Tk::Listbox>
 	}
 
 	sub yview {
-#?		my ($w) = @_;
-#?		$w->Callback(-updatecommand=>$w->can('Tk::Listbox::yview'),@_);     
 		my ($w) = @_;  #JWT:NOTE:Callback->can() EATS 1ST ARGUMENT, SO WE MUST *NOT* SHIFT @_!!!
 		@args = @_;
 		shift @args;   #JWT:NOTE:WE DO SHIFT THE MODULE OFF BEFORE CALLING yview, ETC. DIRECTLY!
@@ -784,8 +819,8 @@ L<Tk::MListbox> L<Tk::Listbox>
 				-pixelwidth => ['PASSIVE'],
 				-bitmap => [qw/SELF bitmap bitmap/, 'noarrow'],
 				-compound => [qw/SELF compound compound/, 'right'],
-				-background    => [qw/SELF background Background/, $Tk::NORMAL_BG],
-				-foreground    => [qw/SELF foreground Foreground/, $Tk::NORMAL_FG]
+				-background    => [qw/SELF background Background/, undef],
+				-foreground    => [qw/SELF foreground Foreground/, undef]
 		);
 	}
 
@@ -816,6 +851,7 @@ L<Tk::MListbox> L<Tk::Listbox>
 		$w->SUPER::Populate($args);
 		my $hdrBG = $args->{'-headerbackground'} || $args->{'-background'} || undef;
 		my $hdrFG = $args->{'-headerforeground'} || $args->{'-foreground'} || undef;
+		my $disableFG = $args->{'-disabledforeground'} || undef;
 
 		## SMLColumn Components
 		## $sep - separator - Frame
@@ -825,7 +861,8 @@ L<Tk::MListbox> L<Tk::Listbox>
 
 		my $sep = $w->Component(
 				Frame   => 'separator',
-				-height => 1
+				-height => 1,
+				-takefocus => 0,
 		)->pack(qw/-side right -fill y -anchor w/);
 
 		$sep->bind( "<B1-Motion>", 
@@ -834,10 +871,12 @@ L<Tk::MListbox> L<Tk::Listbox>
 		[$w=>'Callback','-configurecommand']);
 
 		my $f = $w->Component(
-				Frame => "frame"
+				Frame => "frame",
+				-takefocus => 0,
 		)->pack(qw/-side left -anchor n -fill both -expand 1/);
 
-		my $hdr;		if (defined $undln) { 
+		my $hdr;
+		if (defined $undln) { 
 		
 			$hdr = $f->SMHButton(
 					-takefocus=>0,
@@ -867,15 +906,15 @@ L<Tk::MListbox> L<Tk::Listbox>
 #DECIDED TO LEAVE UNDERLINED:				-activestyle => 'dotbox',
 		)->pack(qw/-side top -anchor n -expand 1 -fill both/);
 		$w->Advertise("listbox" => $lb);
-
 		$w->Delegates (DEFAULT => $lb);
 
-		my $bgWidgets = $hdrBG ? [$f, $lb] : [$f, $hdr, $lb];
-		my $fgWidgets = $hdrFG ? [$f, $lb] : [$f, $hdr, $lb];
+		my $bgWidgets = defined($hdrBG) ? [$f, $lb] : [$f, $hdr, $lb];
+		my $fgWidgets = defined($hdrFG) ? [$f, $lb] : [$f, $hdr, $lb];
+		my $dfgWidgets = defined($disableFG) ? [$f, $lb] : [$f, $hdr, $lb];
 		$w->ConfigSpecs(
 #				-background     => [[$f, $hdr, $lb], 
 				-background     => [$bgWidgets,
-						qw/background Background/, $Tk::NORMAL_BG],
+						qw/background Background/, undef],
 				-comparecommand => ['CALLBACK', undef, undef,
 						sub{$_[0] cmp $_[1]}],
 				-configurecommand => ['CALLBACK'],
@@ -883,13 +922,14 @@ L<Tk::MListbox> L<Tk::Listbox>
 				-font           => [[$hdr, $lb], qw/font Font/, undef],
 #				-foreground     => [[$hdr, $lb],
 				-foreground     => [$fgWidgets,
-						qw/foreground Foreground/, $Tk::NORMAL_FG],
+						qw/foreground Foreground/, undef],
 				-headerbackground => [qw/PASSIVE headerBackground HeaderBackground/, $hdrBG],
 				-headerforeground => [qw/PASSIVE headerForeground HeaderForeground/, $hdrFG],
+				-disabledforeground => [$dfgWidgets, qw/disabledForeground DisabledForeground/, $disableFG],
 				-separatorwidth => [{-width => $sep}, 
 						qw/separatorWidth Separator 1/],
 				-separatorcolor => [{-background => $sep}, 
-						qw/sepaatorColor Separator black/],
+						qw/separatorColor Separator black/],
 				-resizeable     => [qw/METHOD resizeable Resizeable 1/],
 				-sortable       => [qw/PASSIVE sortable Sortable 1/],
 				-text           => [$hdr],
@@ -897,6 +937,7 @@ L<Tk::MListbox> L<Tk::Listbox>
 				-updatecommand  => [$lb],
 				-textwidth      => [{-width => [$lb, $hdr]}],
 				-reversearrow   => [qw/PASSIVE reversearrow reversearrow 0/],
+				-state          => ['METHOD', 'state',  'State',   'normal'],
 				DEFAULT         => [$lb]
 		);
 		$w->ConfigAlias(
@@ -904,8 +945,9 @@ L<Tk::MListbox> L<Tk::Listbox>
 				-width      => '-textwidth'
 		);
 
-		$hdr->configure('-background' => $hdrBG)  if ($hdrBG);
-		$hdr->configure('-foreground' => $hdrFG)  if ($hdrFG);
+		$hdr->configure('-background' => $hdrBG)  if (defined $hdrBG);
+		$hdr->configure('-foreground' => $hdrFG)  if (defined $hdrFG);
+		$hdr->configure('-disabledforeground' => $disableFG)  if (defined $disableFG);
 	}
 
 ######################################################################
@@ -914,8 +956,8 @@ L<Tk::MListbox> L<Tk::Listbox>
 
 	sub resizeable {
 		my ($w, $value) = @_;
-		return $w->{Configure}{-resizeable} unless
-		defined $value;
+		return $w->{Configure}{-resizeable}  unless (defined $value);
+
 		$w->Subwidget("separator")->configure(
 				-cursor => ($value ? 'sb_h_double_arrow' : 'xterm')
 		);
@@ -928,8 +970,7 @@ L<Tk::MListbox> L<Tk::Listbox>
 	sub setWidth {
 		my ($w, $pixels) = @_;
 		$pixels -= $w->Subwidget("separator")->width;
-		return 
-		unless $pixels >= 0;
+		return  unless ($pixels >= 0);
 
 		$w->Subwidget("listbox")->GeometryRequest(
 		$pixels,$w->Subwidget("listbox")->height);
@@ -948,8 +989,24 @@ L<Tk::MListbox> L<Tk::Listbox>
 
 ## Adjust size of column.
 	sub adjustMotion {
-		my ($w) = @_;    
-		$w->setWidth($w->pointerx - $w->rootx);
+		my ($w) = @_;
+		return  if ($w->getState() =~ /d/o);
+
+		$w->setWidth($w->pointerx - $w->rootx)  if ($w->resizeable);
+	}
+
+	sub state {
+		my $w = shift;
+		my $state = shift;
+#FOOBARS FOCUS?!		return $w->{Configure}{'-state'} || undef  unless (defined($state) && $state);
+
+		$w->Subwidget("listbox")->configure('-state' => $state, -takefocus => 0);
+		$w->Subwidget("heading")->configure('-state' => $state, -takefocus => 0);
+		$w->{'_hlistboxstate'} = $state;
+	}
+
+	sub getState {
+		return shift->{'_hlistboxstate'};
 	}
 
 } ## END PRELOADING OF SMLColumn
@@ -967,7 +1024,7 @@ package Tk::SMListbox;
 use strict;
 use Carp;
 use vars qw($VERSION);
-$VERSION = '2.13';
+$VERSION = '2.2';
 
 use Tk;
 
@@ -982,15 +1039,16 @@ sub Tk::Widget::Scrolled {
 	my ($parent, $kind, %args) = @_;
 
 	my $colAR;
-	$colAR = delete $args{'-columns'} if $kind eq "SMListbox";
+	$colAR = delete $args{'-columns'} if $kind eq 'SMListbox';
 
 	## Find args that are Frame create time args
 	my @args = Tk::Frame->CreateArgs($parent,\%args);
 	my $name = delete $args{'Name'};
 	push(@args,'Name' => $name) if (defined $name);
 	my $cw = $parent->Frame(@args);
-	$cw->bind('<FocusIn>', sub { $cw->focusNext; Tk->break;});
+#	$cw->bind('<FocusIn>', sub { $cw->focusNext; Tk->break;});
 	@args = ();
+	my $scrollbarfocus = (defined $args{'-scrollbarfocus'}) ? delete($args{'-scrollbarfocus'}) : undef;
 
 	## Now remove any args that Frame can handle
 	foreach my $k ('-scrollbars',map($_->[0],$cw->configure)) {
@@ -998,23 +1056,36 @@ sub Tk::Widget::Scrolled {
 	}
 	## Anything else must be for target widget - pass at widget create time
 	my $w  = $cw->$kind(%args);
-
+	$cw->{'__SMListbox__'} = $w  if ($w =~ /SMListbox/o);  #JWT:SAVE ME FOR FUTURE USE!
 	## Now re-set %args to be ones Frame can handle
 	## RCS NOTE: I've also slightly modified the ConfigSpecs
 	%args = @args;
+	$args{'-scrollbarfocus'} = $scrollbarfocus  if (defined($scrollbarfocus) && $scrollbarfocus =~ /[xy01]/);
 	$cw->ConfigSpecs(
 			'-scrollbars' => [qw/METHOD   scrollbars Scrollbars se/],
 			'-background' => [$w, qw/background Background/, undef],
 			'-foreground' => [$w, qw/foreground Foreground/, undef],
+			'-scrollbarfocus' => ['METHOD', 'scrollbarfocus', 'Focus', ''],
 	);
 	$cw->AddScrollbars($w);
 	$cw->Default("\L$kind" => $w);
 	$cw->Delegates('bind' => $w, 'bindtags' => $w, 'menu' => $w);
 	$cw->ConfigDefault(\%args);
+	delete($args{'-scrollbarfocus'})  if (defined $args{'-scrollbarfocus'});  #DELETE IT AGAIN SO THIS ONE WON'T BITCH!
 	$cw->configure(%args);
+	$args{'-scrollbarfocus'} = $scrollbarfocus  if (defined($scrollbarfocus) && $scrollbarfocus =~ /[xy01]/);
 	$cw->configure(-columns => $colAR) if $colAR;
+	$w->scrollbarfocus($scrollbarfocus)  if ($scrollbarfocus =~ /[xy01]/);
+	#MAKE HORIZ. SCROLLBAR SCROLL WINDOW HORIZONTALLY (SCROLLING IN COLUMNS SCROLLS JUST THAT COLUMN):
+	$w->parent->Subwidget('xscrollbar')->bind('<Button-6>', sub {
+		$cw->{'__SMListbox__'}->_chgView('xview','scroll',-1,'units')  if (defined $cw->{'__SMListbox__'});
+	});
+	$w->parent->Subwidget('xscrollbar')->bind('<Button-7>', sub {
+		$cw->{'__SMListbox__'}->_chgView('xview','scroll',1,'units')  if (defined $cw->{'__SMListbox__'});
+	});
 	return $cw;
 }
+
 use warnings;
 
 require Tk::Pane;
@@ -1037,15 +1108,17 @@ sub ClassInit {
 	$mw->bind($class,'<Control-slash>','_selectAll');
 	$mw->bind($class,'<Control-backslash>','_deselectAll');
 	$mw->bind($class,'<FocusIn>','focus');
+	$mw->bind($class,'<FocusOut>','unfocus');
 	$mw->bind($class,'<Escape>', '_Cancel'); 
-	$mw->bind($class, '<Home>',  ['_chgView','xview','moveto',0]);
+	$mw->bind($class, '<Home>',  ['_chgView','xview','moveto',0.001]);
 	$mw->bind($class, '<End>',   ['_chgView','xview','moveto',1]);
 	$mw->bind($class, '<Prior>',  ['_chgView','yview','scroll',-1,'pages']);
 	$mw->bind($class, '<Next>',   ['_chgView','yview','scroll',1,'pages']);
 	$mw->bind($class, '<Control-Prior>', ['CtrlPriorNext',-1]);
 	$mw->bind($class, '<Control-Next>', ['CtrlPriorNext',1]);
- $mw->bind($class,'<space>',['SpaceSelect',Ev('index','active')]);
- $mw->bind($class,'<Select>',['SpaceSelect',Ev('index','active')]);
+	$mw->bind($class,'<space>',['SpaceSelect',Ev('index','active')]);
+	$mw->bind($class,'<Select>',['SpaceSelect',Ev('index','active')]);
+	$mw->bind($class,'<Tab>', sub { my $w = shift; $w->focusNext; });
 
 	my $downArrowBits = pack("b10"x10,
 			"..........",
@@ -1086,7 +1159,6 @@ sub ClassInit {
 			".........."
 	);
 	$mw->DefineBitmap('noarrow' => 10,10, $noArrowBits);
-
 }
 
 ## Do some slightly tricky stuff: The -columns option, if called is
@@ -1120,10 +1192,12 @@ sub Populate {
 	$w->{'_sort_descending'} = 0;   ## Flag for ascending/desc. sort order
 	$w->{'_top'} = 0;
 	$w->{'_bottom'} = 0;
+	$w->{'_hasfocus'} = 0;
 
 	my $pane = $w->Component(
 			Pane => "pane",
-			-sticky => 'nsew'
+			-sticky => 'nsew',
+			-takefocus => 0,
 	)->pack(-expand=>1,-fill=>'both');
 
 	my $font;
@@ -1136,19 +1210,20 @@ sub Populate {
 	my $hdr = $w->Subwidget('heading');
 
 	$w->ConfigSpecs(
-			-background        => [qw/METHOD background Background/, $Tk::NORMAL_BG ],
+			-background        => [qw/METHOD background Background/, undef],
 			-columns           => [qw/METHOD/],
 			-configurecommand  => [qw/CALLBACK/],
 			-font              => [qw/METHOD font Font/, $font],
-			-foreground        => [qw/METHOD foreground Foreground/, $Tk::NORMAL_FG ],
-			-headerbackground  => [qw/METHOD background Background/, $Tk::NORMAL_BG ],
-			-headerforeground  => [qw/METHOD foreground Foreground/, $Tk::NORMAL_FG ],
+			-foreground        => [qw/METHOD foreground Foreground/, undef],
+			-headerbackground  => [qw/METHOD headerbackground headerBackground/, undef],
+			-headerforeground  => [qw/METHOD headerforeground headerForeground/, undef],
 			-height            => [qw/METHOD height Height 10/],
 			-moveable          => [qw/PASSIVE moveable Moveable 1/],
 			-resizeable        => [qw/METHOD resizeable Resizeable 1/],
-			-selectbackground  => [qw/METHOD selectBackground Background/, $Tk::SELECT_BG],
+			-selectbackground  => [qw/METHOD selectBackground Background/, undef],
 			-selectborderwidth => [qw/METHOD selectBorderwidth Borderwidth 1/],
-			-selectforeground  => [qw/METHOD selectForeground Foreground/, $Tk::SELECT_FG],
+			-selectforeground  => [qw/METHOD selectForeground Foreground/, undef],
+			-disabledforeground => [qw/PASSIVE disabledForeground DisabledForeground/, undef],
 			-selectmode        => [qw/METHOD selectMode Mode browse/],
 			-compound          => [qw/METHOD compound compound right/],
 			-showallsortcolumns      => [qw/METHOD showallsortcolumns showallsortcolumns 0/],
@@ -1156,15 +1231,17 @@ sub Populate {
 			-fillcolumn        => [qw/PASSIVE fillcolumn fillColumn -1/],
 			-focuscolumn       => [qw/PASSIVE focuscolumn focusColumn -1/],
 			-nocolumnfocus     => [qw/PASSIVE nocolumnfocus nocolumnfocus 0/],
-			-separatorcolor    => [qw/METHOD separatorColor Separator black/],
-			-separatorwidth    => [qw/METHOD separatorWidth Separator 1/], 
+			-separatorcolor    => [qw/METHOD separatorColor Foreground black/],
+			-separatorwidth    => [qw/METHOD separatorWidth SeparatorWidth 1/], 
 			-sortable          => [qw/METHOD sortable Sortable 1/],
-			-takefocus         => [qw/PASSIVE takeFocus Focus 1/],
+			-takefocus         => ['METHOD', 'takeFocus', 'TakeFocus', ''],  #JWT:MUST BE METHOD SINCE SCROLLED EATS IT OTHERWISE!
 			-textwidth         => [qw/METHOD textWidth Width 10/],
 			-width             => [qw/METHOD width Width/, undef],
 			-xscrollcommand    => [$pane],
 			-yscrollcommand    => ['CALLBACK'],  
+			-state             => ['METHOD',  'state',         'State',   'normal'],
 	);
+	$w->{'_ourtakefocus'} = (defined $args->{'-takefocus'}) ? $args->{'-takefocus'} : '';  #NOTE:  Scrolled EATS THIS!
 
 	$w->ConfigAlias(
 			-selectbg => "-selectbackground",
@@ -1173,6 +1250,89 @@ sub Populate {
 			-sepcolor => "-separatorcolor",
 			-sepwidth => "-separatorwidth",
 	);
+	#THESE NEEDED TO GET TAB-FOCUS CYCLING RIGHT:
+	$w->parent->bind('<Tab>', sub { my $w = shift; $w->focusNext; });
+	$w->parent->bind('<FocusIn>', sub { my $w = shift; $w->focusNext; });
+	#JWT:HAD TO DO THESE THIS WAY INSTEAD OF IN THE CLASS, ELSE THEY DON'T WORK!:
+	$w->bindRows('<Home>',  ['_chgView', 'xview','moveto',0.001]);
+	$w->bindRows('<End>',   ['_chgView', 'xview','moveto',1]);
+	$w->bindRows('<Control-ButtonPress-1>', ['CtrlButtonPress', Ev('index',Ev('@'))]);
+
+	$w->bindRows('<ButtonPress-1>', [sub {
+		my $w = shift;
+		if (Tk::Exists($w)) {
+			return  if ($w->{Configure}{'-state'} =~ /d/o);
+
+			if (!$w->{'_hasfocus'} && $w->{'_ourtakefocus'}) {
+				$w->focus;
+				#OBTAIN INDEX OF ROW MOUSE CLICKED ON (I CAN'T FIGURE OUT WHAT TO DO W/THIS "Ev STUFF"?)! :(
+				my $mousexy = '@'.($w->pointerx-$w->rootx).','.($w->pointery-$w->rooty);
+				my $clickedon = $w->index($mousexy) - 1;  #FOR SOME REASON, IT'S ALWAYS OFF-BY-1!
+				$w->activate($clickedon)  if ($clickedon >= 0);  #EMULATE HMListbox.
+			}
+		}
+	}, Ev('index',Ev('@'))]
+	);
+
+	$w->bindRows('<ButtonRelease-1>', [sub {
+		my $w = shift;
+		my $clickedon = $_[0];
+
+		if (Tk::Exists($w)) {
+			return  if ($w->{Configure}{'-state'} =~ /d/o);
+			$w->focus  if (!$w->{'_hasfocus'} && $w->{'_ourtakefocus'});
+		}
+	}, Ev('index',Ev('@'))]
+	);
+
+	$w->bindRows('<Alt-ButtonPress-1>', [sub {
+		my $w = shift;
+		my $clickedon = $_[0];
+
+		if (Tk::Exists($w)) {
+			return  if ($w->{Configure}{'-state'} =~ /d/o);
+
+			my $index = $w->_firstVisible->index('anchor');
+			$w->focus  if (!$w->{'_hasfocus'} && $w->{'_ourtakefocus'});
+			$w->activate($index);
+		}
+	}, Ev('index',Ev('@'))]
+	);
+}
+
+sub CtrlButtonPress {
+	my $w = shift;
+	return  unless (defined $w->_firstVisible);
+
+	my $clickedon = $w->_firstVisible->index('active');
+	my @cursel = $w->_firstVisible->curselection;
+	if (Tk::Exists($w)) {
+		return  if ($w->{Configure}{'-state'} =~ /d/o);
+
+		$w->activate($w->_firstVisible->index('anchor'));
+	}
+}
+
+sub scrollbarfocus { #JWT:  ALLOW THEM TO SPECIFY WHICH, IF ANY SCROLLBARS TAKE FOCUS:
+	my $w = shift;
+	my $focusopt = shift;
+	return $w->{Configure}{'-scrollbarfocus'} || undef  unless (defined($focusopt));
+
+	if ($focusopt eq '0') {
+		$w->parent->Subwidget('xscrollbar')->configure(-takefocus => 0);
+		$w->parent->Subwidget('yscrollbar')->configure(-takefocus => 0);
+	} elsif (! $focusopt) {
+		return;
+	} elsif ($focusopt eq '1' || ($focusopt =~ /x/o && $focusopt =~ /y/o)) {
+		$w->parent->Subwidget('xscrollbar')->configure(-takefocus => 1);
+		$w->parent->Subwidget('yscrollbar')->configure(-takefocus => 1);
+	} elsif ($focusopt =~ /x/o) {
+		$w->parent->Subwidget('xscrollbar')->configure(-takefocus => 1);
+		$w->parent->Subwidget('yscrollbar')->configure(-takefocus => 0);
+	} elsif ($focusopt =~ /y/o || $focusopt !~ /x/o) {
+		$w->parent->Subwidget('yscrollbar')->configure(-takefocus => 1);
+		$w->parent->Subwidget('xscrollbar')->configure(-takefocus => 0);
+	}
 }
 
 sub getSortOrder {   #JWT:  ADDED THIS METHOD TO FETCH HOW THE LIST IS SORTED!
@@ -1233,6 +1393,7 @@ sub font              { shift->_configureColumns('-font', @_) }
 sub foreground        { shift->_configureColumns('-foreground', @_) }
 sub headerbackground  { shift->_configureColumns('-headerbackground', @_);  }
 sub headerforeground  { shift->_configureColumns('-headerforeground', @_);  }
+sub disabledforeground { shift->_configureColumns('-disabledforeground', @_);  }
 sub height            { shift->_configureColumns('-height', @_) }
 sub resizeable        { shift->_configureColumns('-resizeable', @_) }
 sub selectbackground  { shift->_configureColumns('-selectbackground', @_) }
@@ -1282,6 +1443,14 @@ sub separatorwidth    { shift->_configureColumns('-separatorwidth', @_ ) }
 sub sortable          { shift->_configureColumns('-sortable', @_) }
 sub textwidth         { shift->_configureColumns('-textwidth', @_) }
 
+sub takefocus {
+	my ($w, $val, $byus) = @_;
+	return $w->{Configure}{'-takefocus'}  unless (defined $val);
+
+	$w->{'_ourtakefocus'} = $val  unless (defined $byus);
+	$w->{Configure}{'-takefocus'} = ($val =~ /0/o) ? 0 : '';
+}
+
 sub width {
 	my ($w, $v) = @_;
 
@@ -1320,7 +1489,7 @@ sub _bindCallback {
 }
 
 ## bind subwidgets is used by other public bind methods to
-## apply a callback to an event dequence of a particular subwidget 
+## apply a callback to an event sequence of a particular subwidget 
 ## within each of the columns. Any defined callbacks are passed
 ## to the _bindCallback which is actually the callback that gets
 ## bound. 
@@ -1348,7 +1517,7 @@ sub _bindSubwidgets {
 	my @args = ('_bindCallback', $callback);
 	foreach (@{$w->{'_columns'}}) {
 		my $sw = $_->Subwidget($subwidget);
-		if ($sw->class ne "SMCListbox") {
+		if ($sw->class ne 'SMCListbox') {
 			$sw->Tk::bind($sequence, [$w => @args, $sw, $col++]);
 		} else {
 			$sw->Tk::bind($sequence, [$w => @args, $sw, $col++, Ev('y')]);
@@ -1365,7 +1534,7 @@ sub _bindSubwidgets {
 
 sub _configureColumns {
 	my ($w, $option, $value) = @_;
-	return $w->{Configure}{$option}  unless $value;
+	return $w->{Configure}{$option}  unless (defined $value);
 
 	foreach (@{$w->{'_columns'}}) {
 		$_->configure("$option" => $value);
@@ -1378,6 +1547,8 @@ sub _cntrlHome { shift->_firstVisible->Cntrl_Home; }
 
 sub _chgView { 
 	my ($w, $fn, @args) = @_;
+
+	$fn = shift(@args)  if ($fn =~ /HASH/o);
 	my $code = "\$w->$fn('".join("','",@args)."')";
 	eval $code;
 }
@@ -1406,17 +1577,12 @@ sub _deselectAll {
 
 ## implements sorting and dragging & drop of a column
 sub _dragOrSort {
-	my ($w, $c, $sortAnyway) = @_;	if (defined($sortAnyway) && $sortAnyway) { 
-	
-		if ($c->cget('-sortable')) {
-			$w->sort (undef, $c);
-		}
-		return;
-	}
-	unless ($w->cget('-moveable')) {
-		if ($c->cget('-sortable')) {
-			$w->sort (undef, $c);
-		}
+	my ($w, $c, $sortAnyway) = @_;
+
+	return  if ($w->state() =~ /disable/o);  #DON'T ALLOW SORTING OR DRAGGING IF DISABLED!
+
+	if (!$w->cget('-moveable') || (defined($sortAnyway) && $sortAnyway)) {
+		$w->sort(undef, $c)  if ($c->cget('-sortable'));
 		return;
 	}
 
@@ -1458,7 +1624,10 @@ sub _dragOrSort {
 		});
 
 	$h->bind("<ButtonRelease-1>", sub {
-		my $rootx = $tl->rootx;
+		my $rootx;
+		eval {$rootx = $tl->rootx;};
+		return  unless (defined $rootx);  #ASSUME WE'RE DISABLED.
+
 		my $x = $rootx + ($tl->width/2);
 		$tl->destroy;    # Don't need this anymore...
 		$h->bind("<Motion>",'');  # Cancel binding
@@ -1523,7 +1692,7 @@ sub _getEntryFromY {
 	my ($cw, $sw, $yCoord) = @_;
 	my $nearest = $sw->nearest($yCoord);
 
-	return $nearest  if ($nearest < ($sw->size() - 1));
+	return $nearest  if (defined($nearest) && $nearest < ($sw->size() - 1));
 
 	my ($x, $y, $w, $h) = $sw->bbox($nearest);
 	my $lastY = $y + $h;
@@ -1550,7 +1719,7 @@ sub _selectionUpdate {
 			&$code($_->Subwidget("listbox"), @args);
 		}
 	} else {
-#?		&$code($w->{'_columns'}->[0]->Subwidget("listbox"));
+#		&$code($w->{'_columns'}->[0]->Subwidget("listbox"));
 		foreach (@{$w->{'_columns'}}) {
 			&$code($_->Subwidget("listbox"));
 		}
@@ -1622,31 +1791,27 @@ sub focus
 {
 	my $w = shift;
 
-	if (!$w->cget('-takefocus')) { 
-		$w->focusNext;
-	} else {
-		my $c = (defined($w->{Configure}{'-focuscolumn'}) && $w->{Configure}{'-focuscolumn'} >= 0)
-				? $w->columnGet($w->{Configure}{'-focuscolumn'})  #User specified which one to get focus.
-				: $w->_firstVisible; 
-                  #Default to 1st one visible if user did not pick one.
-		if (defined($c) && $w->cget('-nocolumnfocus') != 1) {
-			$c->Subwidget("listbox")->focus(@_);
-			$c->Subwidget("listbox")->bind('<<LeftTab>>', sub {
-				$w->focusPrev;
-				$w->focusCurrent->focusPrev;
-				Tk->break;
-			});
-		} else {
-			$w->Tk::focus(@_);
-			$w->bind('<<LeftTab>>', sub {
-				$w->focusPrev;
-				$w->focusCurrent->focusPrev;
-				Tk->break;
-			});
-		}
+	$w->Tk::focus;
+	$w->{'_hasfocus'} = 1;
+	my $c = (defined($w->{Configure}{'-focuscolumn'}) && $w->{Configure}{'-focuscolumn'} >= 0)
+			? $w->columnGet($w->{Configure}{'-focuscolumn'})  #User specified which one to get focus.
+			: $w->_firstVisible; 
+              #Default to 1st one visible if user did not pick one.
+	if (defined($c) && $w->cget('-nocolumnfocus') != 1) {
+		$c->Subwidget("listbox")->focus(@_);
+		$c->Subwidget("listbox")->bind('<<LeftTab>>', sub {
+			$w->focusPrev;
+			$w->focusCurrent->focusPrev;
+			Tk->break;
+		});
 	}
 }
 
+sub unfocus
+{
+	shift->{'_hasfocus'} = 0;
+}
+
 sub bindColumns    {  shift->_bindSubwidgets('heading',@_) }
 sub bindRows       {  shift->_bindSubwidgets('listbox',@_) }
 sub bindSeparators {  shift->_bindSubwidgets('separator',@_) }
@@ -1749,9 +1914,9 @@ sub columnInsert {
 
 	## Copy these options from the megawidget.
 	foreach (qw/-background -foreground -headerbackground -headerforeground -font -height 
-			-resizeable -selectbackground -selectforeground 
-			-selectborderwidth -selectmode -separatorcolor
-			-separatorwidth -sortable -textwidth -reversearrow/) 
+			-resizeable -selectbackground -selectforeground -disabledforeground 
+			-selectborderwidth -selectmode -separatorcolor 
+			-separatorwidth -sortable -textwidth -reversearrow -state/) 
 	{
 		$opts{$_} = $w->cget($_) if defined $w->cget($_);
 	}
@@ -1776,6 +1941,7 @@ no warnings;
 use warnings;
 		$w->{Configure}{'-jwtlistboxhack'} = $@ ? 0 : 1;  #eval succeeds if using JWT's hacked listbox!
 	}
+	$c->{'__SMListbox__'} = $w;   #JWT:SAVE ME FOR FUTURE REFERENCES.
 	## RCS: Review this later - questionable implementation
 	## Fill the new column with empty values, making sure all columns have
 	## the same number of rows.
@@ -1784,7 +1950,7 @@ use warnings;
 			$c->insert('end','');
 		}
 	}  
-	$c->Subwidget("heading")->bind("<Button-1>", [ $w => '_dragOrSort', $c]);
+	$c->Subwidget("heading")->bind("<ButtonPress-1>", [ $w => '_dragOrSort', $c]);
 	$c->Subwidget("heading")->configure(-compound => $w->{'-compound'})  if ($w->{'-compound'});
 	my @bindTags = $w->bindtags;	
 	if (defined($hotchar) && $hotchar =~ /\S/o)   { #JWT:Set up Alt-<hotkey> bindings if we have a hotkey character defined:
@@ -1902,9 +2068,14 @@ sub getRow {
 	}
 }
 
-sub index { shift->_firstVisible->index(@_)}
+sub index {
+	my $w = shift;
+	
+	return undef  unless (defined($_[0]) && defined($w->_firstVisible));
+	return $w->_firstVisible->index(@_)
+}
 
-sub insert {
+sub insert {    #NOTE:  UNDERLYING LISTBOXES (UNLIKE HLISTBOXES) WILL *NOT* ACCEPT DATA INSERSIONS IF DISABLED!!!
 	my ($w, $index, @data) = @_;
 	my ($rownum, $colnum);
 	my $rowcnt = $#data;
@@ -2060,6 +2231,43 @@ sub sort {
 	}
 }
 
+sub state {
+	my ($w, $val) = @_;
+
+	return $w->{Configure}{'-state'} || undef  unless (defined($val) && $val);
+	return  if (defined($w->{'_prevstate'}) && $val eq $w->{'_prevstate'});  #DON'T DO TWICE IN A ROW!
+
+	$w->{'_statechg'} = 1;
+	if ($val =~ /d/o) {              #WE'RE DISABLING (SAVE CURRENT ENABLED STATUS STUFF, THEN DISABLE USER-INTERACTION):
+#?		$w->{Configure}{'-state'} = 'normal';
+		$w->{'_saveactive'} = $w->index('active') || 0;
+		$w->{'_foreground'} = $w->cget('-foreground');  #SAVE CURRENT (ENABLED) FG COLOR!
+		$w->{Configure}{'-state'} = $val;
+		foreach my $i (0..$#{$w->{'_columns'}}) {
+			$w->{'_columns'}->[$i]->state($val);
+		}
+#?		$w->activate($w->{'_saveactive'})  unless ($w->{'_hasfocus'});
+		$w->takefocus(0, 1);
+		$w->focusNext  if ($w->{'_hasfocus'});  #MOVE FOCUS OFF WIDGET IF IT HAS IT.
+		#NOTE:  THIS DIFFERS FROM Tk::Listbox BEHAVIOUR, AS Tk::Listbox WILL STILL TAKE FOCUS IN THE
+		#TAB-CIRCULATE ORDER EVEN IF DISABLED, IF -takefocus IS SET TO 1!:
+	} elsif ($w->{'_prevstate'}) {   #WE'RE ENABLING (RESTORE PREV. ENABLED STUFF AND REALLOW USER-INTERACTION):
+		my $fg = $w->{'_foreground'};
+		$w->{Configure}{'-state'} = $val;
+		foreach my $i (0..$#{$w->{'_columns'}}) {
+			$w->{'_columns'}->[$i]->state($val);
+		}
+		if (defined $w->_firstVisible) {   #RESTORE SELECTED LIST AND ACTIVE CURSOR:
+			my @selected = $w->_firstVisible->curselection;
+			$w->selectionSet(shift @selected)  while (@selected);
+			$w->activate($w->{'_saveactive'});
+		}
+		$w->takefocus($w->{'_ourtakefocus'}, 1);
+	}
+	$w->{'_prevstate'} = $w->{Configure}{'-state'};
+	$w->{'_statechg'} = 0;
+}
+
 # Implements horizontal scanning. 
 sub xscan {
 	my ($w, $type, $x) = @_;
@@ -2081,6 +2289,7 @@ sub xscan {
 
 sub xview { shift->Subwidget("pane")->xview(@_) }
 sub yview { shift->_firstVisible->yview(@_)}
-1;
+
+1
 
 __END__
